import numpy as np
import pandas as pd
import warnings
import scipy.stats as st
import anndata
from packaging import version

def diffuse_stepwise(data, s, maxnsteps=15):
    # find connectivity matrix
    av = anndata.__version__
    if type(av) == str:
        av = version.parse(av)
    if av < version.parse("0.7.2"):
        a = data.uns["neighbors"]["connectivities"]
    else:
        a = data.obsp["distances"]

    # normalize
    colsums = np.array(a.sum(axis=0)).flatten() + 1

    # do diffusion
    for i in range(maxnsteps):
        print('\ttaking step', i+1)
        s = a.dot(s/colsums[:,None]) + s/colsums[:,None]
        yield s

def diffuse(data, s, nsteps):
    for s in diffuse_stepwise(data, s, maxnsteps=nsteps):
        pass
    return s

def _df_to_array(data, x):
    if type(x) in [pd.DataFrame, pd.Series]:
        if all(x.index == data.samplem.index):
            return x.values
        else:
            print('ERROR: index does not match index of data.samplem')
    else:
        return x

# creates a neighborhood abundance matrix
def _nam(data, nsteps=None, maxnsteps=15):
    s = pd.get_dummies(data.obs_sampleids)[data.samplem.index.values]
    C = s.sum(axis=0)

    prevmedkurt = np.inf
    for i, s in enumerate(diffuse_stepwise(data, s, maxnsteps=maxnsteps)):
        medkurt = np.median(st.kurtosis(s/C, axis=1))
        print('\tmedian kurtosis:', medkurt+3)
        if nsteps is None:
            if prevmedkurt - medkurt < 3 and i+1 >= 3:
                print('stopping after', i+1, 'steps')
                break
            prevmedkurt = medkurt
        elif i+1 == nsteps:
            break

    snorm = (s / C).T
    snorm.index.name = data.samplem.index.name
    return snorm

def _batch_kurtosis(NAM, batches):
    return st.kurtosis(np.array([
                NAM[batches == b].mean(axis=0) for b in np.unique(batches)
                ]), axis=0) + 3

#qcs a NAM to remove neighborhoods that are batchy
def _qc_nam(NAM, batches):
    N = len(NAM)
    if len(np.unique(batches)) == 1:
        warnings.warn('only one unique batch supplied to qc')
        keep = np.repeat(True, len(NAM.T))
        return NAM, keep

    kurtoses = _batch_kurtosis(NAM, batches)
    threshold = max(6, 2*np.median(kurtoses))
    print('throwing out neighborhoods with batch kurtosis >=', threshold)
    keep = (kurtoses < threshold)
    print('keeping', keep.sum(), 'neighborhoods')

    return NAM[:, keep], keep

# residualizes covariates and batch information out of NAM
def _resid_nam(NAM, covs, batches, ridge=None):
    N = len(NAM)
    NAM_ = NAM - NAM.mean(axis=0)
    if covs is None:
        covs = np.ones((N, 0))
    else:
        covs = (covs - covs.mean(axis=0))/covs.std(axis=0)

    if batches is None or len(np.unique(batches)) == 1:
        warnings.warn('only one unique batch supplied to prep')
        C = covs
        if len(C.T) == 0:
            M = np.eye(N)
        else:
            M = np.eye(N) - C.dot(np.linalg.solve(C.T.dot(C), C.T))
        NAM_ = M.dot(NAM_)
    else:
        B = pd.get_dummies(batches).values
        B = (B - B.mean(axis=0))/B.std(axis=0)
        C = np.hstack([B, covs])

        if ridge is not None:
            ridges = [ridge]
        else:
            ridges = [1e5, 1e4, 1e3, 1e2, 1e1, 1e0, 1e-1, 1e-2, 1e-3, 1e-4, 0]

        for ridge in ridges:
            L = np.diag([1]*len(B.T)+[0]*(len(C.T)-len(B.T)))
            M = np.eye(N) - C.dot(np.linalg.solve(C.T.dot(C) + ridge*len(C)*L, C.T))
            NAM_ = M.dot(NAM_)

            kurtoses = _batch_kurtosis(NAM_, batches)

            print('\twith ridge', ridge, 'median batch kurtosis = ',
                    np.median(kurtoses))

            if np.median(kurtoses) <= 6:
                break

    return NAM_ / NAM_.std(axis=0), M, len(C.T)

#performs SVD of NAM
def _svd_nam(NAM):
    U, svs, UT = np.linalg.svd(NAM.dot(NAM.T))
    V = NAM.T.dot(U) / np.sqrt(svs)

    return (U, svs, V)

def nam(data, batches=None, covs=None, filter_samples=None,
    nsteps=None, max_frac_pcs=0.15, suffix='', ks = None,
    force_recompute=False, **kwargs):
    def safe_same(A, B):
        if A is None: A = np.zeros(0)
        if B is None: B = np.zeros(0)
        if A.shape == B.shape:
            return np.allclose(A, B)
        else:
            return False

    # error checking
    covs = _df_to_array(data, covs)
    batches = _df_to_array(data, batches)
    if batches is None:
        batches = np.array([1]*data.N)
    if filter_samples is None:
        if covs is not None:
            filter_samples = ~np.any(np.isnan(covs), axis=1)
        else:
            filter_samples = np.repeat(True, data.N)
    else:
        filter_samples = _df_to_array(data, filter_samples)

    du = data.uns
    # compute and QC NAM
    if force_recompute or \
        'NAM.T'+suffix not in du or \
        not safe_same(batches, du['_batches'+suffix]):
        print('qcd NAM not found; computing and saving')
        NAM = _nam(data, nsteps=nsteps)
        NAMqc, keep = _qc_nam(NAM.values, batches)
        du['NAM.T'+suffix] = pd.DataFrame(NAMqc, index=NAM.index, columns=NAM.columns[keep]).T
        du['keptcells'+suffix] = keep
        du['_batches'+suffix] = batches

    # correct for batch/covariates and SVD
    if force_recompute or \
        'NAM_sampleXpc'+suffix not in du or \
        not safe_same(covs, du['_covs'+suffix]) or \
        not safe_same(filter_samples, du['_filter_samples'+suffix]):

        print('covariate-adjusted NAM not found; computing and saving')
        du['_filter_samples'+suffix] = filter_samples
        NAM = du['NAM.T'+suffix].T.iloc[filter_samples]
        NAM_resid, M, r = _resid_nam(NAM.values,
                                covs[filter_samples] if covs is not None else covs,
                                batches[filter_samples] if batches is not None else batches)

        print('computing SVD')
        U, svs, V = _svd_nam(NAM_resid)
        npcs = min(V.shape[1], max([10]+[int(max_frac_pcs * data.N)]+[ks if ks is not None else []][0]))
        du['NAM_sampleXpc'+suffix] = pd.DataFrame(U,
            index=NAM.index,
            columns=['PC'+str(i) for i in range(1, len(U.T)+1)])
        du['NAM_svs'+suffix] = svs
        du['NAM_varexp'+suffix] = svs / len(U) / len(V)
        du['NAM_nbhdXpc'+suffix] = pd.DataFrame(V[:,:npcs],
            index=NAM.columns,
            columns=['PC'+str(i) for i in range(1, npcs+1)])
        du['_M'+suffix] = M
        du['_r'+suffix] = r
        du['_covs'+suffix] = (np.zeros(0) if covs is None else covs)



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

def umap_ncorr(data, res, fdr_thresh=None, **kwargs):
    if fdr_thresh is None:
        fdr_thresh = 0.05

    passed = res.fdrs[res.fdrs.fdr <= fdr_thresh]
    if len(passed) == 0:
        print('no neighborhoods were significant at FDR <', fdr_thresh)
        thresh = np.infty
    else:
        thresh = passed.threshold.iloc[0]

    ix1 = np.repeat(False, len(data))
    ix1[res.kept] = np.abs(res.ncorrs) > thresh
    c = res.ncorrs[np.abs(res.ncorrs) > thresh]

    umap_overlay(data, ix1, c, **kwargs)

def umap_overlay(data, ix1, c,
    scatter0={},
    scatter1={},
    ax=None, noframe=True, colorbar=True, **cbar_kw):

    # set default plotting options
    scatter0_ = {'alpha':0.1, 's':2, 'color':'grey'}
    scatter1_ = {'alpha':0.2, 's':8, 'c':c, 'cmap':'seismic',
                    'vmin':-np.abs(c).max() if len(c) > 0 else 0,
                    'vmax':np.abs(c).max() if len(c) > 0 else 1}
    scatter0_.update(scatter0)
    scatter1_.update(scatter1)

    if ax is None:
        ax = plt.gca()

    # do plotting
    ax.scatter(*data.obsm['X_umap'].T, **scatter0_)
    res = ax.scatter(*data.obsm['X_umap'][ix1].T, **scatter1_)

    # remove ticks and spines
    ax.set_xticks([]); ax.set_yticks([])
    if noframe:
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.spines['bottom'].set_visible(False)

    # add colorbar if needed
    if colorbar:
        fig = plt.gcf()
        cbar = fig.colorbar(res, ax=ax, **cbar_kw)
        cbar.set_alpha(1)
        cbar.draw_all()
    else:
        cbar = None

    return ax, cbar



import numpy as np
import scipy.stats as st

def conditional_permutation(B, Y, num):
    """
    Permutes Y conditioned on B num different times.
    """
    batchind = np.array([
        np.where(B == b)[0] for b in np.unique(B)
        ])
    ix = np.concatenate([
        bi[np.argsort(np.random.randn(len(bi), num), axis=0)]
        for bi in batchind
        ])
    bix = np.zeros((len(Y), num)).astype(np.int)
    bix[np.concatenate(batchind)] = ix
    result = Y[bix]
    return Y[bix]

def tail_counts(z, znull, atol=1e-8, rtol=1e-5):
    """
    Computes the number of null z-scores of equal or greater magnitude than each supplied
        z-score, for each null instantiation.

    znull is assumed to be either of shape len(z) or len(z) x k, where k is the number of
        null instantiations.
    """
    # re-shape znull if only one set of null results is supplied
    if len(znull.shape) == 1:
        znull = znull.reshape((-1, 1))

    # square z-scores and prepare to sort/un-sort them
    z2 = z**2
    ix = np.argsort(z2); iix = np.argsort(ix)

    # ask numpy to make a histogram with sorted squared z-scores as boundaries
    bins = np.concatenate([z2[ix] - atol - rtol*z2[ix], [np.inf]])
    hist = np.array([
            np.histogram(zn2, bins=bins)[0]
        for zn2 in znull.T**2])

    # convert histogram into tail counts (in-place)
    tails = np.flip(hist, axis=1)
    np.cumsum(tails, axis=1, out=tails)
    tails = np.flip(tails, axis=1)

    # return tail counts for un-sorted z-scores
    return tails[:, iix]

def empirical_fdrs(z, znull, thresholds):
    """
    znull is assumed to be of shape len(z) x k, where k is the number of
        null instantiations.
    """
    # get tail counts
    if znull.shape[0] != len(z):
        print('ERROR: znull is shape', znull.shape, 'and z is shape', z.shape)
    tails = tail_counts(thresholds, znull)
    ranks = tail_counts(thresholds, z)

    # compute FWER (superceded by empirical_fwers)
    #fwer = ((tails > 0).sum(axis=0) + 1) / (znull.shape[1] + 1)

    # compute FDPs
    fdp = tails / ranks
    fdr = fdp.mean(axis=0)
    #fep95 = np.percentile(fdp, 95, axis=0, interpolation='higher')

    return fdr

def empirical_fwers(z, Nmaxz2, atol=1e-8, rtol=1e-5):
    # Nmaxz2 is assumed to be of length k where k is number of null simulates
    tc = tail_counts(z, np.sqrt(Nmaxz2), atol=atol, rtol=rtol)[0]
    return (tc + 1)/(len(Nmaxz2)+1)

def minfwer_loo(Nmaxz2, atol=1e-8, rtol=1e-5):
    tc = np.array([(Nmaxz2 >= z2).sum() for z2 in Nmaxz2])
    return (tc + 1)/len(Nmaxz2)

def numtests(Nmaxz2):
    j, k = 0, 10
    maxs = np.sort(Nmaxz2)[::-1]
    fwers = (np.arange(j, k)+1)/(len(maxs)+1)
    ps = st.chi2.sf(maxs[j:k], 1)
    return 1/(ps.dot(fwers) / fwers.dot(fwers))

def numtests_loo(Nmaxz2):
    return np.array([
        numtests(Nmaxz2[np.arange(len(Nmaxz2)) != i])
        for i in range(len(Nmaxz2))
        ])
